/*-------------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  temperature-humidity meter by Jack Botner

  main.c - the main program

  This program runs a combined temperature and humidity meter.

  Processor: ATmega164P
      Clock: 8 mHz internal
	Display: 3 digit LCD (simple) LCD-S301C31TR
	Sensor: SHT30 temperature and humidity

  The SHT30 is an I2C device that communicates with the
  microcontroller using the TWI pins.

  The battery voltage is also sampled and applied to the AVR's
  ADC7 pin (via 3:1 voltage divider). It is monitored to make sure
  the battery has not run down.

  TWI support code from Peter Fleury.

  20211022 Adapted from original HDC1080 temperature-humidity project.
-------------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "i2cmaster.h"
#include "sht30.h"

//#define TCNT0_INITIAL_VALUE		6	// so timer/counter0 increments 250 times
#define AREF_1P1_MV			1084L	    // 1.1V Aref, measured (millivolts)

// The following allows you to adjust the output of the adc when you know
// the inaccuracy of your TC1047. Unfortunately it is specified as +/- 2
// degrees C, which is not so good. If you can get an accurate reading of
// your device, you can use the formula in the spec sheet to come up with
// a correction factor. To begin with set the define to 1000L, and change
// if needed.

//#define TEMP_CF_1000s       1000L	// temperature correction factor 1000s

#define MIN_RH					   0
#define MAX_RH				      99

#define STATE_INITIAL			  0     // program states
#define STATE_ADC 				  1
#define STATE_WAIT_2SEC			  2
#define STATE_WAIT_HALFSEC	      3
#define STATE_DECIDE			  5
#define STATE_TEMPERATURE	      6
#define STATE_HUMIDITY			  7

// Global variables

static volatile uint16_t	uiTenthsBatteryVoltage=0;

// Internal Functions

static void display_battery(void);
static void display_humidity( uint16_t uiRh );
static void display_temperature( int16_t );
static void system_init(void);

/*-----------------------------------------------------------------
  main()
-----------------------------------------------------------------*/
int main( void )
{
  int8_t 	    ii;
  uint8_t	    ucState=STATE_INITIAL;
  uint16_t	    uiAdc, uiTv;
  unsigned      uiHumidity;
  unsigned char ucRc;
  int           iTemp1;

  while (1)
  {
    switch ( ucState )
	{
	  case STATE_INITIAL:
		system_init();                	// initialize system
        ucState = STATE_ADC;
	    break;

	    case STATE_ADC:						      // sample battery voltage
  		  ADMUX  = 0b01000111;				      // Vref=AVcc; result right adjusted; select ADC7
  		  ADCSRA = 0b10000110;				      // Enable ADC; Disable interrupt; prescaler = 64
		  set_bit( ADCSRA, ADSC );			      // start ADC conversion
		  while( !test_bit( ADCSRA, ADIF ) );	  // wait until completion
  		  uiAdc = (uint16_t) ADCL | ( (uint16_t) ADCH << 8 );
  		  uiTv = ( uiAdc * 50 ) / 1024;
  		  uiTenthsBatteryVoltage = uiTv * 3;
  		  display_battery();
		  ucState = STATE_WAIT_2SEC;
	      break;

	    case STATE_WAIT_2SEC:
  		  for ( ii=0; ii<60; ++ii )			// 2 sec delay
    	    _delay_loop_2( 65000 );			// 32 ms delay
		  ucState = STATE_DECIDE;           // goto sample selected
		  break;

	    case STATE_DECIDE:
	      if ( test_bit( PINB, PB2 ) )		// sw2 determines what we measure
		      ucState = STATE_TEMPERATURE;
		    else
		      ucState = STATE_HUMIDITY;
	      break;

	    case STATE_TEMPERATURE:				// get and display temperature
        ucRc = sht30_readTempHum( SHT30_READ_TEMP, &iTemp1 );
        if ( ucRc )
        {
          set_bit( DIAG_LED_P, DIAG_LED_B );
          LcdClear();
          LcdOutputNumber( 2, 1 );         // digit 1 = 2
          LcdOutputNumber( ucRc, 3 );      // digit 3 = error code
          ucState = STATE_TEMPERATURE;     // hang
        }
        else
        {
  		  display_temperature( iTemp1 );
		  ucState = STATE_WAIT_HALFSEC;
        }
	    break;

	    case STATE_HUMIDITY:					// get and display humidity
        ucRc = sht30_readTempHum( SHT30_READ_HUMID, (int *) &uiHumidity );
        if ( ucRc )
        {
          set_bit( DIAG_LED_P, DIAG_LED_B );
          LcdClear();
          LcdOutputNumber( 3, 1 );         // digit 1 = 3
          LcdOutputNumber( ucRc, 3 );      // digit 3 = error code
 		  ucState = STATE_HUMIDITY;
        }
        else
        {
  		  display_humidity( uiHumidity );
	      ucState = STATE_WAIT_HALFSEC;
        }
	      break;

	    case STATE_WAIT_HALFSEC:
  		  for ( ii=0; ii<15; ++ii )			// 2 sec delay
    	    _delay_loop_2( 65000 );			// 32 ms delay
		  ucState = STATE_DECIDE;
		  break;
	  }               // end switch
  }                 // end do forever
}

/*-----------------------------------------------------------------
  system_init()

  PORTA: PA0 (input) ADC0 battery voltage sense
		 PA1 (output) LCD "B3" segment
		 PA2 (output) LCD "C3" segment
		 PA3 (output) LCD "D3" segment
		 PA4 (output) LCD "E3" segment
		 PA5 (output) LCD "F3" segment
		 PA6 (output) LCD "G3" segment
		 PA7 (input) ADC7 temperature sensor

  PORTB: PB0 (output) LCD "A3" segment
         PB1 (input) T1 input from humidity sensor
         PB2 (input) SW2 with pull-up
         PB3 (input) Celsius/Fahrenheit jumper with pull-up
         PB4 (output) diagnostic LED
         PB5 (output) MOSI ISP
         PB6 (output) MISO ISP
         PB7 (output) SCK ISP

  PORTC: PC0 (output) LCD "A1" segment
         PC1 (output) LCD "B1" segment
         PC2 (output) LCD "C1" segment
         PC3 (output) LCD "D1" segment
         PC4 (output) LCD "E1" segment
         PC5 (output) LCD "F1" segment
         PC6 (output) LCD "G1" segment
		 PC7 (output) LCD "DP1" segment

  PORTD: PD0 (output) LCD "A2" segment
         PD1 (output) LCD "B2" segment
         PD2 (output) LCD "C2" segment
         PD3 (output) LCD "D2" segment
         PD4 (output) LCD "E2" segment
         PD5 (output) LCD "F2" segment
         PD6 (output) LCD "G2" segment
		 PD7 (output) LCD "DP2" segment
----------------------------------------------------------------*/
static void system_init()
{
  // The fuses are set to provide an internal clock, running at
  // 8.0 MHz with a prescaler of 8. Change the prescaler from
  // 8 to 1. The procedure requires two steps as per the device
  // specifications.
  CLKPR = 0b10000000;       // set clock prescaler change enable
  CLKPR = 0b00000000;       // set division factor to 1

  // Set up the I/O ports directions, pull-ups
  DDRA  = 0b01111111;       // PortA0..6=out, 7=input (ADC)
  PORTA = 0b00000000;       // initialize PortA

  DDRB  = 0b00010011;       // PortB0,1=output, 2,3=input, 4=output, 5..7=input
  PORTB = 0b11101100;       // initialize PortB, PB2,3=pullup

  DDRC  = 0b11111100;       // PortC0,1=in, 2..7=output
  PORTC = 0b00000000;       // initialize PortC

  DDRD  = 0b11111111;       // PortD0..7 = output
  PORTD = 0b00000000;       // initialize PortD

  // Set up the ADC
  ADMUX  = 0b01000000;		// Vref=AVcc; result right adjusted; select ADC0
  ADCSRA = 0b10000110;		// Enable ADC; Disable interrupt; prescaler = 64
  ADCSRB = 0b00000000;		// Auto trigger source = none
  DIDR0  = 0b10000000;		// Disable ADC7 digital input buffers

  // Set up the TWI / I2C
  i2c_init();

  PRR = 0b01111110;		  // power save: using TWI and ADC only
  //sei();                // enable global interrupts
}

/*-----------------------------------------------------------------
  display_temperature(...)

  Format the tenths of degrees and display on LCD.
-----------------------------------------------------------------*/
static void display_temperature( int16_t iTemp10C )
{
  int16_t	 	iLen, iTemp;
  char			cBuffer[6], cDisplay[4];
  uint8_t		ucDec;

  if ( iTemp10C > 899 )				// in the oven?
    return;

  if ( test_bit( PINB, PB3 ) )		// jb20111031 was mistakenly PORTB
	iTemp = iTemp10C;						// report degrees C
  else								
	iTemp = ( ( iTemp10C * 9 ) / 5 ) + 320;	// report degrees F

  cDisplay[0] = cDisplay[1] = 0xff;		// in case of error display 0
  cDisplay[2] = '0';
  ucDec = 0;

  itoa( iTemp, cBuffer, 10 );

  // Format the result and insert decimal
  iLen = strlen( cBuffer );

  if ( iLen )
  {
	if ( cBuffer[0] == '-' )
	{
	  // Process negative temperature
	  if ( iLen == 2 )
	  {
	    // Display " -.n"
		cDisplay[0] = 0xff;
		cDisplay[1] = '-';
	    cDisplay[2] = cBuffer[1];
		ucDec = 2;
	  }
	  else
	    if ( iLen == 3 )
		{
		  // Display "-n.n"
		  cDisplay[0] = '-';
		  cDisplay[1] = cBuffer[1];
		  cDisplay[2] = cBuffer[2];
		  ucDec = 2;
		}
		else
		  if ( iLen == 4 )
		  {
		    // Display "-nn" (dropping the tenths digit)
		    cDisplay[0] = '-';
		    cDisplay[1] = cBuffer[1];
		    cDisplay[2] = cBuffer[2];
		    ucDec = 0;
		  }
	}
	else
	{
	  // Process positive temperature
	  ucDec = 2;
	  if ( iLen == 1 )
	  {
	    // Display " 0.n"
		cDisplay[0] = 0xff;
		cDisplay[1] = '0';
	    cDisplay[2] = cBuffer[0];
	  }
	  else
	    if ( iLen == 2 )
		{
		  // Display " n.n"
		  cDisplay[2] = cBuffer[1];
		  cDisplay[1] = cBuffer[0];
		  cDisplay[0] = 0xff;
		}
		else
		{
		  // Display "nn.n"
		  cDisplay[2] = cBuffer[2];
		  cDisplay[1] = cBuffer[1];
		  cDisplay[0] = cBuffer[0];
		}
	}
  }

  LcdOutputNumber( cDisplay[0], 1 );
  LcdOutputNumber( cDisplay[1], 2 );
  LcdOutputNumber( cDisplay[2], 3 );
  LcdOutputDecimal( ucDec );
}

/*-----------------------------------------------------------------
  display_battery()

  Format the tenths of volts and display on LCD.
-----------------------------------------------------------------*/
static void display_battery()
{
  int16_t	 	iLen;
  char			cBuffer[6];

  utoa( uiTenthsBatteryVoltage, cBuffer, 10 );

  // Insert a decimal point in the result
  iLen = strlen( cBuffer );
  if ( !iLen || iLen > 3 )
    return;				// what the heck?

  if ( iLen == 1 )
  {
    cBuffer[2] = cBuffer[0];
	cBuffer[0] = cBuffer[1] = 0xff;
  }
  else
    if ( iLen == 2 )
	{
	  cBuffer[2] = cBuffer[1];
	  cBuffer[1] = cBuffer[0];
	  cBuffer[0] = 0xff;
	}

  LcdOutputNumber( cBuffer[0], 1 );
  LcdOutputNumber( cBuffer[1], 2 );
  LcdOutputNumber( cBuffer[2], 3 );
  LcdOutputDecimal( 2 );
}

/*-----------------------------------------------------------------
  display_humidity(...)

  Format the percent humidity and display on LCD.
-----------------------------------------------------------------*/
static void display_humidity( uint16_t uiRh )
{
  uint16_t	 	uiLen;
  char			cBuffer[6];

  utoa( uiRh, cBuffer, 10 );

  uiLen = strlen( cBuffer );
  if ( !uiLen || uiLen > 2 )
    return;				// what the heck?

  if ( uiLen == 1 )
  {
    cBuffer[2] = cBuffer[0];
	cBuffer[1] = '0';
  }
  else
    if ( uiLen == 2 )
	{
	  cBuffer[2] = cBuffer[1];
	  cBuffer[1] = cBuffer[0];
	}

  cBuffer[0] = 0xff;
  LcdOutputNumber( cBuffer[0], 1 );
  LcdOutputNumber( cBuffer[1], 2 );
  LcdOutputNumber( cBuffer[2], 3 );
  LcdOutputDecimal( 0 );
}
