/*-------------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  hdc1080.c - SHT30 I2C sensor support functions

  Processor: ATmega164P
      Clock: 8 mHz internal

  2021-10-22 written
-------------------------------------------------------------------*/
#include <avr/io.h>
#include <util/delay_basic.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "sht30.h"
#include "i2cmaster.h"

#define CONV_DIVISOR    65535L      // (2**16) - 1

/*-----------------------------------------------------------------
  sht30_readTempHum(...)

  Read the temperature and humidity.

  Input: 1. 1=Return humidity, 2=return temperature
         2. Pointer to integer to return result

  returns: 0 = OK
           1 = START write failed
           2 = write cmd byte 1 fail
           3 = write cmd byte 2 fail
           4 = START read failed
----------------------------------------------------------------*/
unsigned char sht30_readTempHum( unsigned char ucRequest, int *piResult )
{
  unsigned char ucByte, ucRc, ii, ucBuffer[6];
  unsigned int  uiHumid, uiTemp;
  unsigned long ulVal1;
  long          lVal1;

  // Access the bus and device, output the address
  ucRc = i2c_start( SHT30_ADDRESS_W );
  if ( ucRc == 1 )             // fail
    return 1;

  // Output the command
  ucByte = (unsigned char) ( SHT31_MEAS_MEDREP >> 8 );
  ucRc = i2c_write( ucByte );
  if ( ucRc == 1 )             // fail
  {
    i2c_stop();
    return 2;
  }

  ucByte = (unsigned char) ( SHT31_MEAS_MEDREP & 0xff );
  ucRc = i2c_write( ucByte );
  i2c_stop();
  if ( ucRc == 1 )             // fail
    return 3;

  _delay_loop_2 ( 32768 );         // wait 16ms for conversion

  // Access the bus and device, output the address
  ucRc = i2c_start( SHT30_ADDRESS_R );
  if ( ucRc == 1 )             // fail
  return 4;

  // Need to read 6 bytes: 2 bytes temperature, 1 crc, 2 bytes humidity, crc

  for ( ii=0; ii<6; ++ii )
  {
    ucByte = ( ii == 5 ) ? 0 : 1;     // 0 = read followed by stop, 1 = read followed by ack
    ucBuffer[ii] = i2c_read( ucByte );
  }

  i2c_stop();

  // Should check the CRCs ....

  uiTemp  = (unsigned int) ucBuffer[0] << 8 | (unsigned int) ucBuffer[1];
  uiHumid = (unsigned int) ucBuffer[3] << 8 | (unsigned int) ucBuffer[4];

  if ( ucRequest == SHT30_READ_HUMID )
  {
    // Convert the value received to percent humidity
    // See SHT30 data sheet for this formula

    ulVal1 = ( 100L * (unsigned long) uiHumid ) / CONV_DIVISOR;
    *piResult = (int) (long) ulVal1;
  }
  else
  {
    // Convert the value received to degrees celsius
    // See SHT30 data sheet for this formula
    // except we report 10ths degrees

    lVal1 = ( 1750L * (long) uiTemp ) / CONV_DIVISOR;
    lVal1 -= 450L;
    *piResult = (int) lVal1;
  }

  return 0;
}
